"""

We acquire the python information by running an interrogation script via subprocess trigger. This operation is not
cheap, especially not on Windows. To not have to pay this hefty cost every time we apply multiple levels of
caching.
"""
from __future__ import absolute_import, unicode_literals

import json
import logging
import os
import pipes
import sys
from collections import OrderedDict
from hashlib import sha256

from virtualenv.discovery.py_info import PythonInfo
from virtualenv.info import PY2, PY3
from virtualenv.util.path import Path
from virtualenv.util.six import ensure_text
from virtualenv.util.subprocess import Popen, subprocess
from virtualenv.util.zipapp import ensure_file_on_disk
from virtualenv.version import __version__

_CACHE = OrderedDict()
_CACHE[Path(sys.executable)] = PythonInfo()


def from_exe(cls, app_data, exe, raise_on_error=True, ignore_cache=False):
    """"""
    py_info_cache = _get_py_info_cache(app_data)
    result = _get_from_cache(cls, py_info_cache, app_data, exe, ignore_cache=ignore_cache)
    if isinstance(result, Exception):
        if raise_on_error:
            raise result
        else:
            logging.info("%s", str(result))
        result = None
    return result


def _get_py_info_cache(app_data):
    return None if app_data is None else app_data / "py_info" / __version__


def _get_from_cache(cls, py_info_cache, app_data, exe, ignore_cache=True):
    # note here we cannot resolve symlinks, as the symlink may trigger different prefix information if there's a
    # pyenv.cfg somewhere alongside on python3.4+
    exe_path = Path(exe)
    if not ignore_cache and exe_path in _CACHE:  # check in the in-memory cache
        result = _CACHE[exe_path]
    elif py_info_cache is None:  # cache disabled
        failure, py_info = _run_subprocess(cls, exe, app_data)
        result = py_info if failure is None else failure
    else:  # then check the persisted cache
        py_info = _get_via_file_cache(cls, py_info_cache, app_data, exe_path, exe)
        result = _CACHE[exe_path] = py_info
    # independent if it was from the file or in-memory cache fix the original executable location
    if isinstance(result, PythonInfo):
        result.executable = exe
    return result


def _get_via_file_cache(cls, py_info_cache, app_data, resolved_path, exe):
    key = sha256(str(resolved_path).encode("utf-8") if PY3 else str(resolved_path)).hexdigest()
    py_info = None
    resolved_path_text = ensure_text(str(resolved_path))
    try:
        resolved_path_modified_timestamp = resolved_path.stat().st_mtime
    except OSError:
        resolved_path_modified_timestamp = -1
    data_file = py_info_cache / "{}.json".format(key)
    with py_info_cache.lock_for_key(key):
        data_file_path = data_file.path
        if data_file_path.exists() and resolved_path_modified_timestamp != 1:  # if exists and matches load
            try:
                data = json.loads(data_file_path.read_text())
                if data["path"] == resolved_path_text and data["st_mtime"] == resolved_path_modified_timestamp:
                    logging.debug("get PythonInfo from %s for %s", data_file_path, exe)
                    py_info = cls._from_dict({k: v for k, v in data["content"].items()})
                else:
                    raise ValueError("force close as stale")
            except (KeyError, ValueError, OSError):
                logging.debug("remove PythonInfo %s for %s", data_file_path, exe)
                data_file_path.unlink()  # close out of date files
        if py_info is None:  # if not loaded run and save
            failure, py_info = _run_subprocess(cls, exe, app_data)
            if failure is None:
                file_cache_content = {
                    "st_mtime": resolved_path_modified_timestamp,
                    "path": resolved_path_text,
                    "content": py_info._to_dict(),
                }
                logging.debug("write PythonInfo to %s for %s", data_file_path, exe)
                data_file_path.write_text(ensure_text(json.dumps(file_cache_content, indent=2)))
            else:
                py_info = failure
    return py_info


def _run_subprocess(cls, exe, app_data):
    resolved_path = Path(os.path.abspath(__file__)).parent / "py_info.py"
    with ensure_file_on_disk(resolved_path, app_data) as resolved_path:
        cmd = [exe, str(resolved_path)]
        # prevent sys.prefix from leaking into the child process - see https://bugs.python.org/issue22490
        env = os.environ.copy()
        env.pop("__PYVENV_LAUNCHER__", None)
        logging.debug("get interpreter info via cmd: %s", LogCmd(cmd))
        try:
            process = Popen(
                cmd,
                universal_newlines=True,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                env=env,
            )
            out, err = process.communicate()
            code = process.returncode
        except OSError as os_error:
            out, err, code = "", os_error.strerror, os_error.errno
    result, failure = None, None
    if code == 0:
        result = cls._from_json(out)
        result.executable = exe  # keep original executable as this may contain initialization code
    else:
        msg = "failed to query {} with code {}{}{}".format(
            exe, code, " out: {!r}".format(out) if out else "", " err: {!r}".format(err) if err else ""
        )
        failure = RuntimeError(msg)
    return failure, result


class LogCmd(object):
    def __init__(self, cmd, env=None):
        self.cmd = cmd
        self.env = env

    def __repr__(self):
        def e(v):
            return v.decode("utf-8") if isinstance(v, bytes) else v

        cmd_repr = e(" ").join(pipes.quote(e(c)) for c in self.cmd)
        if self.env is not None:
            cmd_repr += e(" env of {!r}").format(self.env)
        if PY2:
            return cmd_repr.encode("utf-8")
        return cmd_repr

    def __unicode__(self):
        raw = repr(self)
        if PY2:
            return raw.decode("utf-8")
        return raw


def clear(app_data):
    py_info_cache = _get_py_info_cache(app_data)
    if py_info_cache is not None:
        with py_info_cache:
            for filename in py_info_cache.path.iterdir():
                if filename.suffix == ".json":
                    with py_info_cache.lock_for_key(filename.stem):
                        if filename.exists():
                            filename.unlink()
    _CACHE.clear()


___all___ = (
    "from_exe",
    "clear",
    "LogCmd",
)



"""Bootstrap"""
from __future__ import absolute_import, unicode_literals

import logging
import os
import sys
from collections import defaultdict
from contextlib import contextmanager
from copy import copy
from shutil import copy2
from zipfile import ZipFile

from virtualenv.info import IS_ZIPAPP
from virtualenv.util.path import Path
from virtualenv.util.six import ensure_str, ensure_text
from virtualenv.util.subprocess import Popen, subprocess
from virtualenv.util.zipapp import ensure_file_on_disk

from . import BUNDLE_SUPPORT, MAX

BUNDLE_FOLDER = Path(os.path.abspath(__file__)).parent


class WheelDownloadFail(ValueError):
    def __init__(self, packages, for_py_version, exit_code, out, err):
        self.packages = packages
        self.for_py_version = for_py_version
        self.exit_code = exit_code
        self.out = out.strip()
        self.err = err.strip()


def get_wheels(for_py_version, wheel_cache_dir, extra_search_dir, packages, app_data, download):
    # not all wheels are compatible with all python versions, so we need to py version qualify it
    processed = copy(packages)
    # 1. acquire from bundle
    acquire_from_bundle(processed, for_py_version, wheel_cache_dir)
    # 2. acquire from extra search dir
    acquire_from_dir(processed, for_py_version, wheel_cache_dir, extra_search_dir)
    # 3. download from the internet
    if download and processed:
        download_wheel(processed, for_py_version, wheel_cache_dir, app_data)

    # in the end just get the wheels
    wheels = _get_wheels(wheel_cache_dir, packages)
    return {p: next(iter(ver_to_files))[1] for p, ver_to_files in wheels.items()}


def acquire_from_bundle(packages, for_py_version, to_folder):
    for pkg, version in list(packages.items()):
        bundle = get_bundled_wheel(pkg, for_py_version)
        if bundle is not None:
            pkg_version = bundle.stem.split("-")[1]
            exact_version_match = version == pkg_version
            if exact_version_match:
                del packages[pkg]
            if version is None or exact_version_match:
                bundled_wheel_file = to_folder / bundle.name
                if not bundled_wheel_file.exists():
                    logging.debug("get bundled wheel %s", bundle)
                    if IS_ZIPAPP:
                        from virtualenv.util.zipapp import extract

                        extract(bundle, bundled_wheel_file)
                    else:
                        copy2(str(bundle), str(bundled_wheel_file))


def get_bundled_wheel(package, version_release):
    return BUNDLE_FOLDER / (BUNDLE_SUPPORT.get(version_release, {}) or BUNDLE_SUPPORT[MAX]).get(package)


def acquire_from_dir(packages, for_py_version, to_folder, extra_search_dir):
    if not packages:
        return
    for search_dir in extra_search_dir:
        wheels = _get_wheels(search_dir, packages)
        for pkg, ver_wheels in wheels.items():
            stop = False
            for _, filename in ver_wheels:
                dest = to_folder / filename.name
                if not dest.exists():
                    if wheel_support_py(filename, for_py_version):
                        logging.debug("get extra search dir wheel %s", filename)
                        copy2(str(filename), str(dest))
                        stop = True
                else:
                    stop = True
                if stop and packages[pkg] is not None:
                    del packages[pkg]
                    break


def wheel_support_py(filename, py_version):
    name = "{}.dist-info/METADATA".format("-".join(filename.stem.split("-")[0:2]))
    with ZipFile(ensure_text(str(filename)), "r") as zip_file:
        metadata = zip_file.read(name).decode("utf-8")
    marker = "Requires-Python:"
    requires = next((i[len(marker) :] for i in metadata.splitlines() if i.startswith(marker)), None)
    if requires is None:  # if it does not specify a python requires the assumption is compatible
        return True
    py_version_int = tuple(int(i) for i in py_version.split("."))
    for require in (i.strip() for i in requires.split(",")):
        # https://www.python.org/dev/peps/pep-0345/#version-specifiers
        for operator, check in [
            ("!=", lambda v: py_version_int != v),
            ("==", lambda v: py_version_int == v),
            ("<=", lambda v: py_version_int <= v),
            (">=", lambda v: py_version_int >= v),
            ("<", lambda v: py_version_int < v),
            (">", lambda v: py_version_int > v),
        ]:
            if require.startswith(operator):
                ver_str = require[len(operator) :].strip()
                version = tuple((int(i) if i != "*" else None) for i in ver_str.split("."))[0:2]
                if not check(version):
                    return False
                break
    return True


def _get_wheels(from_folder, packages):
    wheels = defaultdict(list)
    for filename in from_folder.iterdir():
        if filename.suffix == ".whl":
            data = filename.stem.split("-")
            if len(data) >= 2:
                pkg, version = data[0:2]
                if pkg in packages:
                    pkg_version = packages[pkg]
                    if pkg_version is None or pkg_version == version:
                        wheels[pkg].append((version, filename))
    for versions in wheels.values():
        versions.sort(
            key=lambda a: tuple(int(i) if i.isdigit() else i for i in a[0].split(".")), reverse=True,
        )
    return wheels


def download_wheel(packages, for_py_version, to_folder, app_data):
    to_download = list(p if v is None else "{}=={}".format(p, v) for p, v in packages.items())
    logging.debug("download wheels %s", to_download)
    cmd = [
        sys.executable,
        "-m",
        "pip",
        "download",
        "--disable-pip-version-check",
        "--only-binary=:all:",
        "--no-deps",
        "--python-version",
        for_py_version,
        "-d",
        str(to_folder),
    ]
    cmd.extend(to_download)
    # pip has no interface in python - must be a new sub-process

    with pip_wheel_env_run("{}.{}".format(*sys.version_info[0:2]), app_data) as env:
        process = Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        out, err = process.communicate()
        if process.returncode != 0:
            raise WheelDownloadFail(packages, for_py_version, process.returncode, out, err)


@contextmanager
def pip_wheel_env_run(version, app_data):
    env = os.environ.copy()
    env.update(
        {
            ensure_str(k): str(v)  # python 2 requires these to be string only (non-unicode)
            for k, v in {"PIP_USE_WHEEL": "1", "PIP_USER": "0", "PIP_NO_INPUT": "1"}.items()
        }
    )
    with ensure_file_on_disk(get_bundled_wheel("pip", version), app_data) as pip_wheel_path:
        # put the bundled wheel onto the path, and use it to do the bootstrap operation
        env[str("PYTHONPATH")] = str(pip_wheel_path)
        yield env



from __future__ import absolute_import, unicode_literals

from virtualenv.run.app_data import TempAppData

from .base import PluginLoader


class Discovery(PluginLoader):
    """"""


def get_discover(parser, args):
    discover_types = Discovery.entry_points_for("virtualenv.discovery")
    discovery_parser = parser.add_argument_group(
        title="discovery", description="discover and provide a target interpreter"
    )
    discovery_parser.add_argument(
        "--discovery",
        choices=_get_default_discovery(discover_types),
        default=next(i for i in discover_types.keys()),
        required=False,
        help="interpreter discovery method",
    )
    options, _ = parser.parse_known_args(args)
    if options.app_data == "<temp folder>":
        options.app_data = TempAppData()
    if options.clear_app_data:
        options.app_data.clean()
    discover_class = discover_types[options.discovery]
    discover_class.add_parser_arguments(discovery_parser)
    options, _ = parser.parse_known_args(args, namespace=options)
    discover = discover_class(options)
    return discover


def _get_default_discovery(discover_types):
    return list(discover_types.keys())



from __future__ import absolute_import, unicode_literals

import logging
from contextlib import contextmanager

from virtualenv.discovery.cached_py_info import LogCmd
from virtualenv.info import PY3
from virtualenv.seed.embed.base_embed import BaseEmbed
from virtualenv.seed.embed.wheels.acquire import get_bundled_wheel, pip_wheel_env_run
from virtualenv.util.subprocess import Popen
from virtualenv.util.zipapp import ensure_file_on_disk

if PY3:
    from contextlib import ExitStack
else:
    from contextlib2 import ExitStack


class PipInvoke(BaseEmbed):
    def __init__(self, options):
        super(PipInvoke, self).__init__(options)

    def run(self, creator):
        if not self.enabled:
            return
        with self.get_pip_install_cmd(creator.exe, creator.interpreter.version_release_str) as cmd:
            with pip_wheel_env_run(creator.interpreter.version_release_str, self.app_data) as env:
                logging.debug("pip seed by running: %s", LogCmd(cmd, env))
                process = Popen(cmd, env=env)
                process.communicate()
        if process.returncode != 0:
            raise RuntimeError("failed seed with code {}".format(process.returncode))

    @contextmanager
    def get_pip_install_cmd(self, exe, version):
        cmd = [str(exe), "-m", "pip", "-q", "install", "--only-binary", ":all:"]
        if not self.download:
            cmd.append("--no-index")
        pkg_versions = self.package_version()
        for key, ver in pkg_versions.items():
            cmd.append("{}{}".format(key, "=={}".format(ver) if ver is not None else ""))
        with ExitStack() as stack:
            folders = set()
            for context in (ensure_file_on_disk(get_bundled_wheel(p, version), self.app_data) for p in pkg_versions):
                folders.add(stack.enter_context(context).parent)
            for folder in folders:
                cmd.extend(["--find-links", str(folder)])
                cmd.extend(self.extra_search_dir)
            yield cmd



from __future__ import absolute_import, unicode_literals

BUNDLE_SUPPORT = {
    "3.9": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-46.1.3-py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
    "3.8": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-46.1.3-py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
    "3.7": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-46.1.3-py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
    "3.6": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-46.1.3-py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
    "3.5": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-46.1.3-py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
    "3.4": {
        "pip": "pip-19.1.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-43.0.0-py2.py3-none-any.whl",
        "wheel": "wheel-0.33.6-py2.py3-none-any.whl",
    },
    "2.7": {
        "pip": "pip-20.1-py2.py3-none-any.whl",
        "setuptools": "setuptools-44.1.0-py2.py3-none-any.whl",
        "wheel": "wheel-0.34.2-py2.py3-none-any.whl",
    },
}
MAX = "3.9"



"""Bootstrap"""
from __future__ import absolute_import, unicode_literals

import logging
from contextlib import contextmanager
from functools import partial
from threading import Lock, Thread

from virtualenv.info import fs_supports_symlink
from virtualenv.seed.embed.base_embed import BaseEmbed
from virtualenv.seed.embed.wheels.acquire import WheelDownloadFail, get_wheels
from virtualenv.util.path import safe_delete

from .pip_install.copy import CopyPipInstall
from .pip_install.symlink import SymlinkPipInstall


class FromAppData(BaseEmbed):
    def __init__(self, options):
        super(FromAppData, self).__init__(options)
        self.symlinks = options.symlink_app_data
        self.base_cache = self.app_data / "seed-app-data" / "v1.0.1"

    @classmethod
    def add_parser_arguments(cls, parser, interpreter, app_data):
        super(FromAppData, cls).add_parser_arguments(parser, interpreter, app_data)
        can_symlink = app_data.transient is False and fs_supports_symlink()
        parser.add_argument(
            "--symlink-app-data",
            dest="symlink_app_data",
            action="store_true" if can_symlink else "store_false",
            help="{} symlink the python packages from the app-data folder (requires seed pip>=19.3)".format(
                "" if can_symlink else "not supported - "
            ),
            default=False,
        )

    def run(self, creator):
        if not self.enabled:
            return
        base_cache = self.base_cache / creator.interpreter.version_release_str
        with self._get_seed_wheels(creator, base_cache) as name_to_whl:
            pip_version = name_to_whl["pip"].stem.split("-")[1] if "pip" in name_to_whl else None
            installer_class = self.installer_class(pip_version)

            def _install(name, wheel):
                logging.debug("install %s from wheel %s via %s", name, wheel, installer_class.__name__)
                image_folder = base_cache.path / "image" / installer_class.__name__ / wheel.stem
                installer = installer_class(wheel, creator, image_folder)
                if not installer.has_image():
                    installer.build_image()
                installer.install(creator.interpreter.version_info)

            threads = list(Thread(target=_install, args=(n, w)) for n, w in name_to_whl.items())
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()

    @contextmanager
    def _get_seed_wheels(self, creator, base_cache):
        with base_cache.lock_for_key("wheels"):
            wheels_to = base_cache.path / "wheels"
            if wheels_to.exists():
                safe_delete(wheels_to)
            wheels_to.mkdir(parents=True, exist_ok=True)
            name_to_whl, lock, fail = {}, Lock(), {}

            def _get(package, version):
                wheel_loader = partial(
                    get_wheels,
                    creator.interpreter.version_release_str,
                    wheels_to,
                    self.extra_search_dir,
                    {package: version},
                    self.app_data,
                )
                failure, result = None, None
                # fallback to download in case the exact version is not available
                for download in [True] if self.download else [False, True]:
                    failure = None
                    try:
                        result = wheel_loader(download)
                        if result:
                            break
                    except Exception as exception:
                        failure = exception
                if failure:
                    if isinstance(failure, WheelDownloadFail):
                        msg = "failed to download {}".format(package)
                        if version is not None:
                            msg += " version {}".format(version)
                        msg += ", pip download exit code {}".format(failure.exit_code)
                        output = failure.out + failure.err
                        if output:
                            msg += "\n"
                            msg += output
                    else:
                        msg = repr(failure)
                    logging.error(msg)
                    with lock:
                        fail[package] = version
                else:
                    with lock:
                        name_to_whl.update(result)

            package_versions = self.package_version()
            threads = list(Thread(target=_get, args=(pkg, v)) for pkg, v in package_versions.items())
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
            if fail:
                raise RuntimeError("seed failed due to failing to download wheels {}".format(", ".join(fail.keys())))
            yield name_to_whl

    def installer_class(self, pip_version):
        if self.symlinks and pip_version:
            # symlink support requires pip 19.3+
            pip_version_int = tuple(int(i) for i in pip_version.split(".")[0:2])
            if pip_version_int >= (19, 3):
                return SymlinkPipInstall
        return CopyPipInstall

    def __unicode__(self):
        base = super(FromAppData, self).__unicode__()
        msg = ", via={}, app_data_dir={}".format("symlink" if self.symlinks else "copy", self.base_cache.path)
        return base[:-1] + msg + base[-1]



from __future__ import absolute_import, unicode_literals

from abc import ABCMeta

from six import add_metaclass

from virtualenv.util.path import Path
from virtualenv.util.six import ensure_str, ensure_text

from ..seeder import Seeder


@add_metaclass(ABCMeta)
class BaseEmbed(Seeder):
    packages = ["pip", "setuptools", "wheel"]

    def __init__(self, options):
        super(BaseEmbed, self).__init__(options, enabled=options.no_seed is False)
        self.download = options.download
        self.extra_search_dir = [i.resolve() for i in options.extra_search_dir if i.exists()]

        def latest_is_none(key):
            value = getattr(options, key)
            return None if value == "latest" else value

        self.pip_version = latest_is_none("pip")
        self.setuptools_version = latest_is_none("setuptools")
        self.wheel_version = latest_is_none("wheel")

        self.no_pip = options.no_pip
        self.no_setuptools = options.no_setuptools
        self.no_wheel = options.no_wheel
        self.app_data = options.app_data.folder

        if not self.package_version():
            self.enabled = False

    def package_version(self):
        return {
            package: getattr(self, "{}_version".format(package))
            for package in self.packages
            if getattr(self, "no_{}".format(package)) is False
        }

    @classmethod
    def add_parser_arguments(cls, parser, interpreter, app_data):
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--no-download",
            "--never-download",
            dest="download",
            action="store_false",
            help="pass to disable download of the latest {} from PyPI".format("/".join(cls.packages)),
            default=True,
        )
        group.add_argument(
            "--download",
            dest="download",
            action="store_true",
            help="pass to enable download of the latest {} from PyPI".format("/".join(cls.packages)),
            default=False,
        )
        parser.add_argument(
            "--extra-search-dir",
            metavar="d",
            type=Path,
            nargs="+",
            help="a path containing wheels the seeder may also use beside bundled (can be set 1+ times)",
            default=[],
        )
        for package in cls.packages:
            parser.add_argument(
                "--{}".format(package),
                dest=package,
                metavar="version",
                help="{} version to install, bundle for bundled".format(package),
                default="latest",
            )
        for package in cls.packages:
            parser.add_argument(
                "--no-{}".format(package),
                dest="no_{}".format(package),
                action="store_true",
                help="do not install {}".format(package),
                default=False,
            )

    def __unicode__(self):
        result = self.__class__.__name__
        result += "("
        if self.extra_search_dir:
            result += "extra_search_dir={},".format(", ".join(ensure_text(str(i)) for i in self.extra_search_dir))
        result += "download={},".format(self.download)
        for package in self.packages:
            if getattr(self, "no_{}".format(package)):
                continue
            result += " {}{},".format(
                package, "={}".format(getattr(self, "{}_version".format(package), None) or "latest")
            )
        return result[:-1] + ")"

    def __repr__(self):
        return ensure_str(self.__unicode__())
